module Lecture5 where

import Data.Char

-------------------------------------------------------
--DEMO 1--

data ID = ID Integer deriving (Show, Eq)

data Employee = Employee
  { idNumber :: ID
  , name :: String
  , supervisor :: Maybe ID
  } deriving (Show, Eq)

database :: [Employee]
database =
--           id     name              supervisor
  [ Employee (ID 1) "Jane C. Founder" Nothing
  , Employee (ID 2) "Robert Sysadmin" (Just (ID 1))
  , Employee (ID 3) "Simon Singh"     (Just (ID 1))
  , Employee (ID 4) "Obvious E. Rorr" (Just (ID 5))
  ]

dbLookup :: ID -> Maybe Employee
dbLookup i = case filter (\x -> idNumber x == i) database of
  (e:_) -> Just e
  []    -> Nothing

parseID :: String -> Maybe ID
parseID [] = Nothing
parseID xs
  | all isNumber xs = Just . ID . read $ xs
  | otherwise       = Nothing


-- return the supervisor of the employee with the given ID
-- assuming that the employee exists, the employee has a supervisor,
-- and finally that the supervisor is valid.
getSupervisor :: String -> Maybe Employee
getSupervisor userInput =
  case parseID userInput of
    Nothing -> Nothing
    Just i -> case dbLookup i of
      Nothing -> Nothing
      Just e  -> case supervisor e of
        Nothing -> Nothing
        Just si -> dbLookup si

-- recurring pattern:
--  either we get Nothing and return Nothing
--  or else we get something (i, e, si) and then continue with that
--
{-
       variable part: what we're testing
       vvvvvvvvvvvvvvvvv
  case parseID userInput of
    Nothing -> Nothing -- return Nothing
    Just i ->  rest_of_the_program i -- the rest of the program is
                                     -- a function of i

               ^^^^^^^^^^^^^^^^^^^ the rest of the program
-}

bindM :: Maybe a -> (a -> Maybe b) -> Maybe b
bindM value rest =
  case value of
    Nothing -> Nothing
    Just v  -> rest v

{-
bindM' :: Maybe a -> (a -> Maybe b) -> Maybe b
bindM' Nothing rest = Nothing
bindM' (Just v) rest = rest v
-}
{-

getSupervisor :: String -> Maybe Employee
getSupervisor userInput =
  case parseID userInput of
    Nothing -> Nothing
    Just i -> case dbLookup i of
      Nothing -> Nothing
      Just e  -> case supervisor e of
        Nothing -> Nothing
        Just si -> dbLookup si

-}

getSupervisorB :: String -> Maybe Employee
getSupervisorB userInput =
  bindM (parseID userInput) $ \i  ->
  bindM (dbLookup i)        $ \e  ->
  bindM (supervisor e)      $ \si ->
  dbLookup si



-------------------------------------------------------
--DEMO 2--

data Tree a
  = Leaf
  | Node a (Tree a) (Tree a)
  deriving (Show, Eq)

{-
labeling the tree
    o
   / \
  o   o
     / \
    o   o

should yield
    1
   / \
  2   3
     / \
    4   5
-}

exampleTree :: Tree ()
exampleTree = Node () (Node () Leaf Leaf)
                      (Node () (Node () Leaf Leaf)
                               (Node () Leaf Leaf)
                      )

slideTree :: Tree ()
slideTree = Node () (Node () (Node () Leaf Leaf) (Node () Leaf Leaf))
                    (Node () (Node () Leaf Leaf) (Node () Leaf Leaf))

label :: Tree a -> Tree Integer
label tree = snd (go tree 1) where
  go :: Tree a -> Integer -> (Integer, Tree Integer)
  go Leaf         = \lun -> (lun, Leaf)
  go (Node v l r) = \lun ->
    let (lun', l') = go l (lun + 1) in
    let (lun'', r'') = go r lun' in
    (lun'', Node lun l' r'')

{-                state       state
                  vvvvvvv     vvvvvvv
  type signature: Integer -> (Integer, Tree Integer)
                                       ^^^^^^^^^^^^
                                       return value
  -- final return value: Tree Integer
  -- a global state that we maintain is an Integer value

  \origState ->
    let (newState, retVal1) = ... in
    let (newerState, retVal2) = ... in
    let (newestState, retVal3) = ... in
    ...
-}

type State s a = s -> (s, a)
-- a: eventual return value
-- s: state

bindS :: State s a -> (a -> State s b) -> s -> (s, b)
bindS f rest = \origState ->
  let (newState, retVal) = f origState in
  rest retVal newState


-- helpful State combinators

-- returns the state without changing it
get :: State s s
get = \origState -> (origState, origState)

-- replace the state with the argument (update the state)
put :: s -> State s ()
put newState = \origState -> (newState, ())

-- the function that returns its argument, and leaves the state unchanged
yield :: a -> State s a
yield v = \origState -> (origState, v)

-- use gets the least unused number, and returns it as a value for
-- the rest of the program
-- and also, it will update the least unused number to a larger one.
use :: State Integer Integer
use = \lun -> (lun + 1, lun)

use' :: State Integer Integer
use' =
  bindS get              $ \lun ->
  bindS (put (lun + 1))  $ \()  ->
  yield lun



{-

label :: Tree a -> Tree Integer
label tree = snd (go tree 1) where
  go :: Tree a -> Integer -> (Integer, Tree Integer)
  go Leaf         = \lun -> (lun, Leaf)
  go (Node v l r) = \lun ->
    let (lun', l') = go l (lun + 1) in
    let (lun'', r'') = go r lun' in
    (lun'', Node lun l' r'')

-}

label' :: Tree a -> Tree Integer
label' tree = snd (go tree 1) where
  go :: Tree a -> State Integer (Tree Integer)
  go Leaf = yield Leaf
  go (Node x l r) =
    bindS use     $ \lun ->
    bindS (go l)  $ \l'  ->
    bindS (go r)  $ \r'' ->
    yield (Node lun l' r'')


{-
bindM :: Maybe   a -> (a -> Maybe b)   -> Maybe b
bindS :: State s a -> (a -> State s b) -> State s b

bind  :: m a -> (a -> m b)       -> m b


-}


tryNumber :: Int -> Maybe Int
tryNumber x
  | even x = Just x
  | otherwise = Nothing

toString :: Int -> String
toString = show

maybeMap :: (a -> b) -> Maybe a -> Maybe b
maybeMap f x = case x of
  Nothing -> Nothing
  Just v  -> Just (f v)

treeMap :: (a -> b) -> Tree a -> Tree b
treeMap f x = case x of
  Leaf -> Leaf
  (Node v l r) -> Node (f v) (treeMap f l) (treeMap f r)


type Fun x = Integer -> x

funMap :: (a -> b) -> (Integer -> a) -> Integer -> b
funMap f g x = f (g x)
